{% import 'keg-elements/forms/horizontal-b3.html' as dynamic_render %}
{% import 'keg-elements/forms/horizontal-b4.html' as b4_render %}
{% import 'keg-elements/forms/horizontal-static.html' as static_render %}

<div id="dynamic">
    {{ dynamic_render.form(form) }}
</div>

<div id="b4">
    {{ b4_render.form(form) }}
</div>

<div id="static">
    {{ static_render.form(form) }}
</div>
