{% if is_read_only %}
    {% import 'keg-elements/forms/horizontal-static.html' as horizontal %}
{% else %}
    {% import 'keg-elements/forms/horizontal-b4.html' as horizontal %}
{% endif %}

{% set base_template = config.get('BASE_TEMPLATE') or config.get('KEG_BASE_TEMPLATE') %}
{% if base_template %}
    {% extends base_template %}
{% endif %}

{% block scripts %}
{{ super() }}
{% if config.get('KEG_USE_SELECT2', True) %}
    {% include 'keg-elements/forms/select2-scripts.html' %}
{% endif %}
{{ horizontal.custom_js() }}
{% endblock %}

{% block styles %}
{{ super() }}
{% if config.get('KEG_USE_SELECT2', True) %}
    {% include 'keg-elements/forms/select2-styles.html' %}
{% endif %}
{{ horizontal.custom_css() }}
{% endblock %}

{% block page_content %}
    {% block page_content_title %}
        {% if title %}
        <h1>{{ title }}</h1>
        {% endif %}
    {% endblock %}
    {% block form %}{% endblock %}
{% endblock %}